# This module exists to create the "best" dispatch object for a given
# object.  If "makepy" support for a given object is detected, it is
# used, otherwise a dynamic dispatch object.

# Note that if the unknown dispatch object then returns a known
# dispatch object, the known class will be used.  This contrasts
# with dynamic.Dispatch behaviour, where dynamic objects are always used.

import sys

import pycomint
import pywinint

from . import dynamic
from pywinint import IID

#where are these normally initilized? manually set here -dz
pycomint.IID_NULL = IID("{00000000-0000-0000-0000-000000000000}")
pycomint.IID_IDispatch = IID("{00020400-0000-0000-C000-000000000046}")
pycomint.IID_IUnknown = IID("{00000000-0000-0000-C000-000000000046}")
pycomint.IID_IEnumVARIANT =  IID("{00020404-0000-0000-C000-000000000046}")

_PyIDispatchType = pycomint.TypeIIDs[pycomint.IID_IDispatch]


def __WrapDispatch(
    dispatch,
    userName=None,
    resultCLSID=None,
    typeinfo=None,
    UnicodeToString=None,
    clsctx=pycomint.CLSCTX_SERVER,
    WrapperClass=None,
):

    if WrapperClass is None:
        WrapperClass = CDispatch
    return dynamic.Dispatch(dispatch, userName, WrapperClass, typeinfo, clsctx=clsctx)

def CreateObject(progid):
    #just a shortcut thats more intutive for vb devs
    return Dispatch(progid)
    
def GetObject(Pathname=None, Class=None, clsctx=None):
    """
    Mimic VB's GetObject() function.

    ob = GetObject(Class = "ProgID") or GetObject(Class = clsid) will
    connect to an already running instance of the COM object.

    ob = GetObject(r"c:\blah\blah\foo.xls") (aka the COM moniker syntax)
    will return a ready to use Python wrapping of the required COM object.

    Note: You must specifiy one or the other of these arguments. I know
    this isn't pretty, but it is what VB does. Blech. If you don't
    I'll throw ValueError at you. :)

    This will most likely throw pythoncom.com_error if anything fails.
    """
        
    if clsctx is None:
        clsctx = pycomint.CLSCTX_ALL

    if (Pathname is None and Class is None) or (
        Pathname is not None and Class is not None
    ):
        raise ValueError(
            "You must specify a value for Pathname or Class, but not both."
        )

    if Class is not None:
        return GetActiveObject(Class, clsctx)
    else:
        #first check to see if its an in process COM object shared from VB before we check ROT
        v = pycomint.HostResolver(Pathname); 
        #print(type(v)) #<class 'PyIDispatch'> or <class 'NoneType'>
        if v is not None:
            return __WrapDispatch(v, Pathname, clsctx=clsctx)
        else:
            return Moniker(Pathname, clsctx)


def GetActiveObject(Class, clsctx=pycomint.CLSCTX_ALL):
    """
    Python friendly version of GetObject's ProgID/CLSID functionality.
    """
    resultCLSID = pywintypes.IID(Class)
    dispatch = pycomint.GetActiveObject(resultCLSID)
    dispatch = dispatch.QueryInterface(pycomint.IID_IDispatch)
    return __WrapDispatch(dispatch, Class, resultCLSID=resultCLSID, clsctx=clsctx)


def Moniker(Pathname, clsctx=pycomint.CLSCTX_ALL):
    """
    Python friendly version of GetObject's moniker functionality.
    """
    moniker, i, bindCtx = pycomint.MkParseDisplayName(Pathname)
    dispatch = moniker.BindToObject(bindCtx, None, pycomint.IID_IDispatch)
    return __WrapDispatch(dispatch, Pathname, clsctx=clsctx)


def Dispatch(
    dispatch,
    userName=None,
    resultCLSID=None,
    typeinfo=None,
    UnicodeToString=None,
    clsctx=pycomint.CLSCTX_SERVER,
):
    """Creates a Dispatch based COM object."""
    assert UnicodeToString is None, "this is deprecated and will go away"
    dispatch, userName = dynamic._GetGoodDispatchAndUserName(dispatch, userName, clsctx)
    return __WrapDispatch(dispatch, userName, resultCLSID, typeinfo, clsctx=clsctx)

class CDispatch(dynamic.CDispatch):
    """
    The dynamic class used as a last resort.
    The purpose of this overriding of dynamic.CDispatch is to perpetuate the policy
    of using the makepy generated wrapper Python class instead of dynamic.CDispatch
    if/when possible.
    """

    def _wrap_dispatch_(
        self, ob, userName=None, returnCLSID=None, UnicodeToString=None
    ):
        assert UnicodeToString is None, "this is deprecated and will go away"
        return Dispatch(ob, userName, returnCLSID, None)

    def __dir__(self):
        return dynamic.CDispatch.__dir__(self)

class Constants:
    """A container for generated COM constants."""

    def __init__(self):
        self.__dicts__ = []  # A list of dictionaries

    def __getattr__(self, a):
        for d in self.__dicts__:
            if a in d:
                return d[a]
        raise AttributeError(a)


# And create an instance.
constants = Constants()


# A helpers for DispatchWithEvents - this becomes __setattr__ for the
# temporary class.
def _event_setattr_(self, attr, val):
    try:
        # Does the COM object have an attribute of this name?
        self.__class__.__bases__[0].__setattr__(self, attr, val)
    except AttributeError:
        # Otherwise just stash it away in the instance.
        self.__dict__[attr] = val


############################################
# The base of all makepy generated classes
############################################
class DispatchBaseClass:
    def __init__(self, oobj=None):
        if oobj is None:
            oobj = pycomint.new(self.CLSID)
        elif isinstance(oobj, DispatchBaseClass):
            try:
                oobj = oobj._oleobj_.QueryInterface(
                    self.CLSID, pycomint.IID_IDispatch
                )  # Must be a valid COM instance
            except pycomint.com_error as details:
                from . import winerror

                # Some stupid objects fail here, even tho it is _already_ IDispatch!!??
                # Eg, Lotus notes.
                # So just let it use the existing object if E_NOINTERFACE
                if details.hresult != winerror.E_NOINTERFACE:
                    raise
                oobj = oobj._oleobj_
        self.__dict__["_oleobj_"] = oobj  # so we dont call __setattr__

    def __dir__(self):
        lst = (
            list(self.__dict__.keys())
            + dir(self.__class__)
            + list(self._prop_map_get_.keys())
            + list(self._prop_map_put_.keys())
        )
        try:
            lst += [p.Name for p in self.Properties_]
        except AttributeError:
            pass
        return list(set(lst))

    # Provide a prettier name than the CLSID
    def __repr__(self):
        # Need to get the docstring for the module for this class.
        try:
            mod_doc = sys.modules[self.__class__.__module__].__doc__
            if mod_doc:
                mod_name = "pycom.gen_py." + mod_doc
            else:
                mod_name = sys.modules[self.__class__.__module__].__name__
        except KeyError:
            mod_name = "pycom.unknown"
        return "<%s.%s instance at 0x%s>" % (
            mod_name,
            self.__class__.__name__,
            id(self),
        )

    # Delegate comparison to the oleobjs, as they know how to do identity.
    def __eq__(self, other):
        other = getattr(other, "_oleobj_", other)
        return self._oleobj_ == other

    def __ne__(self, other):
        other = getattr(other, "_oleobj_", other)
        return self._oleobj_ != other

    def _ApplyTypes_(self, dispid, wFlags, retType, argTypes, user, resultCLSID, *args):
        return self._get_good_object_(
            self._oleobj_.InvokeTypes(dispid, 0, wFlags, retType, argTypes, *args),
            user,
            resultCLSID,
        )

    def __getattr__(self, attr):
        args = self._prop_map_get_.get(attr)
        if args is None:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (repr(self), attr)
            )
        return self._ApplyTypes_(*args)

    def __setattr__(self, attr, value):
        if attr in self.__dict__:
            self.__dict__[attr] = value
            return
        try:
            args, defArgs = self._prop_map_put_[attr]
        except KeyError:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (repr(self), attr)
            )
        self._oleobj_.Invoke(*(args + (value,) + defArgs))

    def _get_good_single_object_(self, obj, obUserName=None, resultCLSID=None):
        return _get_good_single_object_(obj, obUserName, resultCLSID)

    def _get_good_object_(self, obj, obUserName=None, resultCLSID=None):
        return _get_good_object_(obj, obUserName, resultCLSID)


# XXX - These should be consolidated with dynamic.py versions.
def _get_good_single_object_(obj, obUserName=None, resultCLSID=None):
    if _PyIDispatchType == type(obj):
        return Dispatch(obj, obUserName, resultCLSID)
    return obj


def _get_good_object_(obj, obUserName=None, resultCLSID=None):
    if obj is None:
        return None
    elif isinstance(obj, tuple):
        obUserNameTuple = (obUserName,) * len(obj)
        resultCLSIDTuple = (resultCLSID,) * len(obj)
        return tuple(map(_get_good_object_, obj, obUserNameTuple, resultCLSIDTuple))
    else:
        return _get_good_single_object_(obj, obUserName, resultCLSID)


class CoClassBaseClass:
    def __init__(self, oobj=None):
        if oobj is None:
            oobj = pycomint.new(self.CLSID)
        dispobj = self.__dict__["_dispobj_"] = self.default_interface(oobj)
        # See comments below re the special methods.
        for maybe in [
            "__call__",
            "__str__",
            "__int__",
            "__iter__",
            "__len__",
            "__nonzero__",
        ]:
            if hasattr(dispobj, maybe):
                setattr(self, maybe, getattr(self, "__maybe" + maybe))

    def __repr__(self):
        return "<pycom.gen_py.%s.%s>" % (__doc__, self.__class__.__name__)

    def __getattr__(self, attr):
        d = self.__dict__["_dispobj_"]
        if d is not None:
            return getattr(d, attr)
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if attr in self.__dict__:
            self.__dict__[attr] = value
            return
        try:
            d = self.__dict__["_dispobj_"]
            if d is not None:
                d.__setattr__(attr, value)
                return
        except AttributeError:
            pass
        self.__dict__[attr] = value

        # Special methods don't use __getattr__ etc, so explicitly delegate here.
        # Note however, that not all are safe to let bubble up - things like
        # `bool(ob)` will break if the object defines __int__ but then raises an
        # attribute error - eg, see #1753.
        # It depends on what the wrapped COM object actually defines whether these
        # will exist on the underlying object, so __init__ explicitly checks if they
        # do and if so, wires them up.

    def __maybe__call__(self, *args, **kwargs):
        return self.__dict__["_dispobj_"].__call__(*args, **kwargs)

    def __maybe__str__(self, *args):
        return self.__dict__["_dispobj_"].__str__(*args)

    def __maybe__int__(self, *args):
        return self.__dict__["_dispobj_"].__int__(*args)

    def __maybe__iter__(self):
        return self.__dict__["_dispobj_"].__iter__()

    def __maybe__len__(self):
        return self.__dict__["_dispobj_"].__len__()

    def __maybe__nonzero__(self):
        return self.__dict__["_dispobj_"].__nonzero__()


# A very simple VARIANT class.  Only to be used with poorly-implemented COM
# objects.  If an object accepts an arg which is a simple "VARIANT", but still
# is very pickly about the actual variant type (eg, isn't happy with a VT_I4,
# which it would get from a Python integer), you can use this to force a
# particular VT.
class VARIANT(object):
    def __init__(self, vt, value):
        self.varianttype = vt
        self._value = value

    # 'value' is a property so when set by pycomint it gets any magic wrapping
    # which normally happens for result objects
    def _get_value(self):
        return self._value

    def _set_value(self, newval):
        self._value = _get_good_object_(newval)

    def _del_value(self):
        del self._value

    value = property(_get_value, _set_value, _del_value)

    def __repr__(self):
        return "pycom.client.VARIANT(%r, %r)" % (self.varianttype, self._value)
